--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- Mock Turtle
-- https://gitlab.cern.ch/coht/mockturtle
--------------------------------------------------------------------------------
--
-- unit name:   spec_mt_demo
--
-- description: Top entity for Mock Turtle SPEC demo.
--
-- This demo instantiates a Mock Turtle CPU with two cores, each connected to
-- a WB GPIO peripheral. Reading and writing to the GPIOs controls the LEDs
-- and buttons of the SPEC.
--
--------------------------------------------------------------------------------
-- Copyright (c) 2018-2019 CERN (home.cern)
--------------------------------------------------------------------------------
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 2.0 (the "License"); you may not use this file except
-- in compliance with the License. You may obtain a copy of the License at
-- http://solderpad.org/licenses/SHL-2.0.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.gn4124_core_pkg.all;
use work.mock_turtle_pkg.all;
use work.mt_mqueue_pkg.all;
use work.synthesis_descriptor.all;

library unisim;
use unisim.vcomponents.all;

entity spec_mt_demo is
  port (
    -- resets and clocks
    clk_125m_pllref_p_i : in    std_logic;
    clk_125m_pllref_n_i : in    std_logic;
    -- SPEC front panel LEDs
    fp_led_green_o      : out   std_logic;
    fp_led_red_o        : out   std_logic;
    -- SPEC board LEDs and buttons
    brd_button_i        : in    std_logic_vector(1 downto 0);
    brd_led_o           : out   std_logic_vector(3 downto 0);
    -- Gennum PCI interface
    gn_rst_n_i          : in    std_logic;
    gn_p2l_clk_n_i      : in    std_logic;
    gn_p2l_clk_p_i      : in    std_logic;
    gn_p2l_rdy_o        : out   std_logic;
    gn_p2l_dframe_i     : in    std_logic;
    gn_p2l_valid_i      : in    std_logic;
    gn_p2l_data_i       : in    std_logic_vector(15 downto 0);
    gn_p_wr_req_i       : in    std_logic_vector(1 downto 0);
    gn_p_wr_rdy_o       : out   std_logic_vector(1 downto 0);
    gn_rx_error_o       : out   std_logic;
    gn_l2p_clkn_o       : out   std_logic;
    gn_l2p_clkp_o       : out   std_logic;
    gn_l2p_dframe_o     : out   std_logic;
    gn_l2p_valid_o      : out   std_logic;
    gn_l2p_edb_o        : out   std_logic;
    gn_l2p_data_o       : out   std_logic_vector(15 downto 0);
    gn_l2p_rdy_i        : in    std_logic;
    gn_l_wr_rdy_i       : in    std_logic_vector(1 downto 0);
    gn_p_rd_d_rdy_i     : in    std_logic_vector(1 downto 0);
    gn_tx_error_i       : in    std_logic;
    gn_vc_rdy_i         : in    std_logic_vector(1 downto 0);
    gn_gpio_b           : inout std_logic_vector(1 downto 0));
end spec_mt_demo;

architecture arch of spec_mt_demo is

  constant c_MT_CONFIG : t_mt_config := (
    app_id     => x"d331d331",
    cpu_count  => 2,
    cpu_config => (others =>
                   (memsize => 8192,
                    hmq_config => (2, (0      => (7, 3, 2, x"0000_0000", false),
                                       1      => (5, 4, 3, x"0000_0000", false),
                                       others => (c_DUMMY_MT_MQUEUE_SLOT))),
                    rmq_config => (1, (0      => (7, 2, 2, x"0000_0000", false),
                                       others => (c_DUMMY_MT_MQUEUE_SLOT))))),
    shared_mem_size => 2048);

  constant c_NUM_WB_MASTERS : integer := 1;
  constant c_NUM_WB_SLAVES  : integer := 2;

  constant c_MASTER_PCIE : integer := 0;

  constant c_SLAVE_VIC      : integer := 0;
  constant c_SLAVE_MT       : integer := 1;
  constant c_DESC_SYNTHESIS : integer := 2;
  constant c_DESC_REPO_URL  : integer := 3;

  constant c_INTERCONNECT_LAYOUT : t_sdb_record_array(c_NUM_WB_SLAVES + 1 downto 0) := (
    c_SLAVE_VIC      => f_sdb_embed_device(c_XWB_VIC_SDB, x"00002000"),
    c_SLAVE_MT       => f_sdb_embed_device(c_MOCK_TURTLE_SDB, x"00020000"),
    c_DESC_SYNTHESIS => f_sdb_embed_synthesis(c_SDB_SYNTHESIS_INFO),
    c_DESC_REPO_URL  => f_sdb_embed_repo_url(c_SDB_REPO_URL));

  constant c_SDB_ADDRESS : t_wishbone_address := x"00000000";

  -- not really used, will be reprogrammed by software
  constant c_VIC_VECTOR_TABLE : t_wishbone_address_array(0 to 3) := (
    0 => x"00020000",   -- MT Mqueue in interrupt
    1 => x"00020001",   -- MT Mqueue out interrupt
    2 => x"00020002",   -- MT Console interrupt
    3 => x"00020003");  -- MT Notify interrupt

  signal cnx_master_out : t_wishbone_master_out_array(c_NUM_WB_MASTERS-1 downto 0);
  signal cnx_master_in  : t_wishbone_master_in_array(c_NUM_WB_MASTERS-1 downto 0);

  signal cnx_slave_out : t_wishbone_slave_out_array(c_NUM_WB_SLAVES-1 downto 0);
  signal cnx_slave_in  : t_wishbone_slave_in_array(c_NUM_WB_SLAVES-1 downto 0);

  signal clk_125m_pllref : std_logic;
  signal clk_sys         : std_logic;
  signal clk_sys_locked  : std_logic;

  signal rst_sys_n     : std_logic;
  signal rstlogic_arst : std_logic;
  signal gn_rst_n_p    : std_logic;

  signal pllout_clk_sys       : std_logic;
  signal pllout_clk_fb_pllref : std_logic;

  signal dp_wb_out : t_wishbone_master_out_array(1 downto 0);
  signal dp_wb_in  : t_wishbone_master_in_array(1 downto 0);

  signal cpu0_gpio_out, cpu1_gpio_out : std_logic_vector(7 downto 0);
  signal cpu_gpio_out, cpu_gpio_in    : std_logic_vector(7 downto 0);

  signal mt_hmq_in_irq  : std_logic;
  signal mt_hmq_out_irq : std_logic;
  signal mt_console_irq : std_logic;
  signal mt_notify_irq  : std_logic;
  signal vic_master_irq : std_logic;

begin  -- architecture arch

  U_Buf_CLK_PLL : IBUFGDS
    generic map (
      DIFF_TERM    => TRUE,
      IBUF_LOW_PWR => TRUE)
    port map (
      O  => clk_125m_pllref,
      I  => clk_125m_pllref_p_i,
      IB => clk_125m_pllref_n_i);

  cmp_sys_clk_pll : PLL_BASE
    generic map (
      BANDWIDTH          => "OPTIMIZED",
      CLK_FEEDBACK       => "CLKFBOUT",
      COMPENSATION       => "INTERNAL",
      DIVCLK_DIVIDE      => 1,
      CLKFBOUT_MULT      => 8,
      CLKFBOUT_PHASE     => 0.000,
      CLKOUT0_DIVIDE     => 16,  -- 62.5 MHz
      CLKOUT0_PHASE      => 0.000,
      CLKOUT0_DUTY_CYCLE => 0.500,
      CLKIN_PERIOD       => 8.0,
      REF_JITTER         => 0.016)
    port map (
      CLKFBOUT => pllout_clk_fb_pllref,
      CLKOUT0  => pllout_clk_sys,
      LOCKED   => clk_sys_locked,
      RST      => '0',
      CLKFBIN  => pllout_clk_fb_pllref,
      CLKIN    => clk_125m_pllref);

  cmp_clk_sys_buf : BUFG
    port map (
      O => clk_sys,
      I => pllout_clk_sys);

  -- Detect when gn_rst_n_i goes high (end of reset) and use this edge to
  -- generate rstlogic_arst. This is needed to connect optional PCIe
  -- reset. When board runs standalone, we need to ignore PCIe reset being
  -- constantly low.
  cmp_arst_edge: gc_sync_ffs
    generic map (
      g_sync_edge => "positive")
    port map (
      clk_i    => clk_sys,
      rst_n_i  => '1',
      data_i   => gn_rst_n_i,
      npulse_o => gn_rst_n_p);

  -- logic AND of all (possibly) async reset sources (active high)
  rstlogic_arst <= (not clk_sys_locked) and gn_rst_n_p;

  cmp_rstlogic_reset : gc_reset_multi_aasd
    generic map (
      g_CLOCKS  => 1,
      g_RST_LEN => 16)
    port map (
      arst_i     => rstlogic_arst,
      clks_i(0)  => clk_sys,
      rst_n_o(0) => rst_sys_n);

  U_Intercon : xwb_sdb_crossbar
    generic map (
      g_NUM_MASTERS => c_NUM_WB_MASTERS,
      g_NUM_SLAVES  => c_NUM_WB_SLAVES,
      g_REGISTERED  => TRUE,
      g_WRAPAROUND  => TRUE,
      g_LAYOUT      => c_INTERCONNECT_LAYOUT,
      g_SDB_ADDR    => c_SDB_ADDRESS)
    port map (
      clk_sys_i => clk_sys,
      rst_n_i   => rst_sys_n,
      slave_i   => cnx_master_out,
      slave_o   => cnx_master_in,
      master_i  => cnx_slave_out,
      master_o  => cnx_slave_in);

  U_VIC : xwb_vic
    generic map (
      g_INTERFACE_MODE      => PIPELINED,
      g_ADDRESS_GRANULARITY => BYTE,
      g_NUM_INTERRUPTS      => 4,
      g_INIT_VECTORS        => c_VIC_VECTOR_TABLE)
    port map (
      clk_sys_i    => clk_sys,
      rst_n_i      => rst_sys_n,
      slave_i      => cnx_slave_in(c_SLAVE_VIC),
      slave_o      => cnx_slave_out(c_SLAVE_VIC),
      irqs_i(0)    => mt_hmq_in_irq,
      irqs_i(1)    => mt_hmq_out_irq,
      irqs_i(2)    => mt_console_irq,
      irqs_i(3)    => mt_notify_irq,
      irq_master_o => vic_master_irq);

  cmp_gn4124_core : xwb_gn4124_core
    port map (
      rst_n_a_i         => gn_rst_n_i,
      status_o          => open,
      p2l_clk_p_i       => gn_p2l_clk_p_i,
      p2l_clk_n_i       => gn_p2l_clk_n_i,
      p2l_data_i        => gn_p2l_data_i,
      p2l_dframe_i      => gn_p2l_dframe_i,
      p2l_valid_i       => gn_p2l_valid_i,
      p2l_rdy_o         => gn_p2l_rdy_o,
      p_wr_req_i        => gn_p_wr_req_i,
      p_wr_rdy_o        => gn_p_wr_rdy_o,
      rx_error_o        => gn_rx_error_o,
      vc_rdy_i          => gn_vc_rdy_i,
      l2p_clk_p_o       => gn_l2p_clkp_o,
      l2p_clk_n_o       => gn_l2p_clkn_o,
      l2p_data_o        => gn_l2p_data_o,
      l2p_dframe_o      => gn_l2p_dframe_o,
      l2p_valid_o       => gn_l2p_valid_o,
      l2p_edb_o         => gn_l2p_edb_o,
      l2p_rdy_i         => gn_l2p_rdy_i,
      l_wr_rdy_i        => gn_l_wr_rdy_i,
      p_rd_d_rdy_i      => gn_p_rd_d_rdy_i,
      tx_error_i        => gn_tx_error_i,
      dma_irq_o         => open,
      irq_p_i           => vic_master_irq,
      irq_p_o           => gn_gpio_b(1),
      wb_master_clk_i   => clk_sys,
      wb_master_rst_n_i => rst_sys_n,
      wb_master_i       => cnx_master_in(c_MASTER_PCIE),
      wb_master_o       => cnx_master_out(c_MASTER_PCIE));

  -- drive unused GN GPIO output
  gn_gpio_b(0) <= '0';

  U_Mock_Turtle : mock_turtle_core
    generic map (
      g_CONFIG            => c_MT_CONFIG,
      g_WITH_WHITE_RABBIT => FALSE)
    port map (
      clk_i         => clk_sys,
      rst_n_i       => rst_sys_n,
      dp_master_o   => dp_wb_out,
      dp_master_i   => dp_wb_in,
      host_slave_i  => cnx_slave_in(c_SLAVE_MT),
      host_slave_o  => cnx_slave_out(c_SLAVE_MT),
      hmq_in_irq_o  => mt_hmq_in_irq,
      hmq_out_irq_o => mt_hmq_out_irq,
      notify_irq_o  => mt_notify_irq,
      console_irq_o => mt_console_irq);

  U_GPIO_CPU0 : xwb_gpio_port
    generic map (
      g_INTERFACE_MODE         => PIPELINED,
      g_ADDRESS_GRANULARITY    => BYTE,
      g_NUM_PINS               => 8,
      g_WITH_BUILTIN_SYNC      => FALSE,
      g_WITH_BUILTIN_TRISTATES => FALSE)
    port map (
      clk_sys_i  => clk_sys,
      rst_n_i    => rst_sys_n,
      slave_i    => dp_wb_out(0),
      slave_o    => dp_wb_in(0),
      gpio_out_o => cpu0_gpio_out,
      gpio_in_i  => cpu_gpio_in,
      gpio_oen_o => open);

  U_GPIO_CPU1 : xwb_gpio_port
    generic map (
      g_INTERFACE_MODE         => PIPELINED,
      g_ADDRESS_GRANULARITY    => BYTE,
      g_NUM_PINS               => 8,
      g_WITH_BUILTIN_SYNC      => FALSE,
      g_WITH_BUILTIN_TRISTATES => FALSE)
    port map (
      clk_sys_i  => clk_sys,
      rst_n_i    => rst_sys_n,
      slave_i    => dp_wb_out(1),
      slave_o    => dp_wb_in(1),
      gpio_out_o => cpu1_gpio_out,
      gpio_in_i  => cpu_gpio_in,
      gpio_oen_o => open);

  cpu_gpio_out <= cpu0_gpio_out or cpu1_gpio_out;

  cpu_gpio_in(7 downto 2) <= cpu_gpio_out(7 downto 2);
  cpu_gpio_in(1 downto 0) <= brd_button_i;

  fp_led_green_o <= cpu_gpio_out(7);
  fp_led_red_o   <= cpu_gpio_out(6);
  brd_led_o      <= cpu_gpio_out(5 downto 2);

end architecture arch;
